//---------------------------------------------------------------------------
#ifndef commonH
#define commonH
//---------------------------------------------------------------------------
#define PROG_NAME             "MakePsn"
#define PROG_EXE_NAME         "makepsn"
#define PROG_VERSION          0x11
#define PROG_AUTHOR           "RM-TEAM 08/2016, 12/2016"

//---------------------------------------------------------------------------
#define BYTE                  unsigned char
#define WORD                  unsigned short int
#define DWORD                 unsigned int
#define BOOL                  int
#define FALSE                 0
#define TRUE                  !FALSE

//---------------------------------------------------------------------------
#define PAGE_SIZE             16384
#define BUFFER_SIZE           65536

//---------------------------------------------------------------------------
#define ERR_OK                0
#define ERR_PRINT_USAGE       -1
#define ERR_OUT_OF_MEMORY     -2
#define ERR_BAD_OPTIONS       -3
#define ERR_NO_OPTION_VALUE   -4
#define ERR_OPEN_SRC_FILE     -5
#define ERR_ACCESS_SRC_FILE   -6
#define ERR_LENGTH_SRC_FILE   -7
#define ERR_SRC_FILE_TO_LONG  -8
#define ERR_READ_SRC_FILE     -9
#define ERR_OPEN_DEST_FILE    -10
#define ERR_ACCESS_DEST_FILE  -11
#define ERR_WRITE_DEST_FILE   -12
#define ERR_BAD_ADDRESS       -13
#define ERR_NO_INPUT_FILES    -14
#define ERR_NO_OUTPUT_FILE    -15
#define ERR_OUTPUT_FILE_MORE  -16
#define ERR_ROM_FILE_MORE     -17
#define ERR_PSN_FILE_MORE     -18
#define ERR_BAD_COMP_MODEL    -19
#define ERR_BAD_REG_NAME      -20
#define ERR_BAD_PAGE_NUM      -21
#define ERR_BAD_MM_VALUE      -22
#define ERR_BAD_VALUE         -23
#define ERR_SNAP_BAD_HEADER   -24
#define ERR_SNAP_UNK_VERSION  -25
#define ERR_SNAP_UNK_MODEL    -26
#define ERR_SNAP_BAD_MEM_BLK  -27
#define ERR_INVALID_HEXSTRING -28
#define ERR_HEXSTRING_TO_LONG -29

//---------------------------------------------------------------------------
#define CM_UNKNOWN            -1
#define CM_V1                 0
#define CM_V2                 1
#define CM_V2A                2
#define CM_V3                 3
#define CM_MATO               4
#define CM_ALFA               5
#define CM_ALFA2              6
#define CM_C2717              7
#define CM_LAST               CM_C2717

//---------------------------------------------------------------------------
#define SNAP_SIGNATURE        "PSN"
#define SNAP_SIGNATURE_LEN    3

#define SNAP_VERSION          2
#define SNAP_HEADER_V1_LEN    56
#define SNAP_HEADER_V2_LEN    124

#define SNAP_MIN_HEADER_LEN   SNAP_HEADER_V1_LEN
#define SNAP_MAX_HEADER_LEN   SNAP_HEADER_V2_LEN
#define SNAP_BLOCK_LEN        16384

#if PAGE_SIZE != SNAP_BLOCK_LEN
  #error "PAGE_SIZE != SNAP_BLOCK_LEN"
#endif

//---------------------------------------------------------------------------
#pragma pack(push, 1)
typedef struct Snap
{
  char cSignature[3];     // "PSN"
  BYTE btVersion;
  WORD wBlocksOffset;
  BYTE btModel;
  BYTE btInterrupt;        // CPU
  WORD wAF;
  WORD wBC;
  WORD wDE;
  WORD wHL;
  WORD wPC;
  WORD wSP;
  WORD wRomLength;        // ROM
  WORD wBlock0Length;     // 0000:3FFF
  WORD wBlock1Length;     // 4000:7FFF
  WORD wBlock2Length;     // 8000:BFFF
  WORD wBlock3Length;     // C000:FFFF
  BYTE btSystemPioCwr;    // System PIO
  BYTE btSystemPioC;
  BYTE btSystemPioA;
  BYTE btGpioCwr;         // GPIO
  BYTE btGpioC;
  BYTE btGpioB;
  BYTE btGpioA;
  BYTE btGpioInt;
  BYTE btIms2Cwr;         // IMS-2
  BYTE btIms2C;
  BYTE btIms2B;
  BYTE btIms2A;
  BYTE btIms2Int;
  BYTE btTimer0Cwr;       // Timer 0
  BYTE btTimer0L;
  BYTE btTimer0H;
  BYTE btTimer1Cwr;       // Timer 1
  BYTE btTimer1L;
  BYTE btTimer1H;
  BYTE btTimer2Cwr;       // Timer 2
  BYTE btTimer2L;
  BYTE btTimer2H;
  BYTE btUsartCwr;        // USART
  BYTE btUsartSync1;
  BYTE btUsartSync2;
  BYTE btUsartCmd;

#if (SNAP_VERSION > 1)
  // verzia 2
  BYTE btVideoIntVGA;
  BYTE btMemMap;
  WORD wBlock4Length;
  WORD wBlock5Length;
  WORD wBlock6Length;
  WORD wBlock7Length;
  WORD wBlock8Length;
  WORD wBlock9Length;
  WORD wBlock10Length;
  WORD wBlock11Length;
  WORD wBlock12Length;
  WORD wBlock13Length;
  WORD wBlock14Length;
  WORD wBlock15Length;
  BYTE btMif85Int;
  BYTE btaMif85Regs[32];
  BYTE btMusicaTimer0Cwr; // Timer 0
  BYTE btMusicaTimer0L;
  BYTE btMusicaTimer0H;
  BYTE btMusicaTimer1Cwr; // Timer 1
  BYTE btMusicaTimer1L;
  BYTE btMusicaTimer1H;
  BYTE btMusicaTimer2Cwr; // Timer 2
  BYTE btMusicaTimer2L;
  BYTE btMusicaTimer2H;
#endif // (SNAP_VERSION > 1)

} SNAP;
#pragma pack(pop)

//---------------------------------------------------------------------------
typedef struct tagDataFile {
  char *pszFn;        // pointer na cestu/meno k suboru
  int iAddress;       // cielova adresa
  int iPage;          // cislo stranky
  struct tagDataFile *pNext;
} DATA_FILE;

//---------------------------------------------------------------------------
typedef struct WorkParams {
  char *pszFnOut;       // pointer na cestu/meno ku vystupnemu suboru
  char *pszFnPsn;       // pointer na cestu/meno "sablony" PSN
  char *pszFnRom;       // pointer na cestu/meno suboru s ROM
  BYTE *pbtRom;         // pointer na buffer s ROM
  BYTE *pbtBuff;        // pointer na pracovny buffer
  SNAP *pSnap;          // pointer na strukturu SNAP
  BYTE *pbtaPages[16];  // pointery na bloky pamati pre stranky
  int iaPgLen[16];      // velkosti dat v bloku pamati
  DATA_FILE *pDFfirst;  // pointer na struktury datovych suborov
  DATA_FILE *pDFlast;   // pointer na struktury datovych suborov
  BOOL bVB;             // verborse print
  BOOL bPack;           // uloz pakovany SNAP
  BOOL bAllram;         // zapnute AllRAM
  BOOL bEI;             // povolene globalne prerusenie
  BOOL bMus;            // zapnutie interface MUSICA
  int iVPI;             // povolenie prerusenia od VideoCPU
  int iRomLen;          // velkost ROM
  int iMM;              // mapovanie pamati 256kB
  int iMif85;           // MIF85 a povolenie prerusenia od CT0
  int iCM;              // model pocitaca
  int iAF;              // hodnota registra AF
  int iBC;              // hodnota registra BC
  int iDE;              // hodnota registra DE
  int iHL;              // hodnota registra HL
  int iPC;              // hodnota registra PC
  int iSP;              // hodnota registra SP
} WORK_PARAMS;

//---------------------------------------------------------------------------
// main.c
int writeFile(char *pszFileName, BYTE *pbtBuff, int iLen, BOOL bAppend);
int packBlock(BYTE *pbtDest, BYTE *pbtSrc, int iLen);
int unpackBlock(BYTE *pbtDest, int iDestlen, BYTE *pbtSrc, int iSrclen);
int allocBuffer(void **ppvBuff, int iSize);
int printError(int iErr, void *pvPar, void *pvPar2);

// snap.c
int loadSnapshot(WORK_PARAMS *pWP);
int saveSnapshot(WORK_PARAMS *pWP);

//---------------------------------------------------------------------------
#endif
//---------------------------------------------------------------------------

 