//---------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "common.h"

//---------------------------------------------------------------------------
int processArguments(int argc, char* argv[], WORK_PARAMS **ppWP);
int fillMemoryPages(WORK_PARAMS *pWP);
int addDataFile(WORK_PARAMS *pWP, char *pszStr, int iPage, int iAddr);
int parseNumber(char *pszStr);
int readFile(char *pszFileName, BYTE *pbtBuff, int iMaxLen);
void version(void);
void usage(void);

//---------------------------------------------------------------------------
int main(int argc, char* argv[])
{
  WORK_PARAMS *pWP = NULL;
  DATA_FILE *pDF1, *pDF2;
  int i, iErr;

  do {
    version();

    if (argc == 1) {
      iErr = ERR_PRINT_USAGE;
      break;
    }

    // spracovanie a kontrola vstupnych parametrov
    iErr = processArguments(argc, argv, &pWP);
    if (iErr != ERR_OK)
      break;

    // vytvorenie bloku SNAP
    iErr = allocBuffer((void *) &pWP->pSnap, sizeof(SNAP));
    if (iErr != ERR_OK)
      break;

    // nacitanie sablony PSN
    if (pWP->pszFnPsn) {
      printf("Loading template Snapshot `%s`\n", pWP->pszFnPsn);
      iErr = loadSnapshot(pWP);
      if (iErr != ERR_OK)
        break;
    }

    // priprav novy obsah pamatovych blokov
    iErr = fillMemoryPages(pWP);
    if (iErr != ERR_OK)
      break;

    // vytvorenie samotneho PSN
    printf("Saving final Snapshot `%s`\n", pWP->pszFnOut);
    iErr = saveSnapshot(pWP);

  } while (FALSE);

  if (iErr == ERR_PRINT_USAGE) {
    usage();
    iErr = 0;
  }
  else if (!iErr)
    printf("\n");

  // * na zaver uvolni alokovanu pamat *

  if (pWP) {
    // stranky pamati
    for (i = 0; i < 16; i++) {
      if (pWP->pbtaPages[i])
        free(pWP->pbtaPages[i]);
    }
    // buffer pre ROM
    if (pWP->pbtRom)
      free(pWP->pbtRom);
    // pracovny buffer
    if (pWP->pbtBuff)
      free(pWP->pbtBuff);
    // struktury
    pDF1 = pWP->pDFfirst;
    while (pDF1) {
      pDF2 = pDF1->pNext;
      free(pDF1);
      pDF1 = pDF2;
    }
    if (pWP->pSnap)
      free(pWP->pSnap);
    free(pWP);
  }

#ifdef _DEBUG
  getchar();
#endif

  return -iErr;
}
//---------------------------------------------------------------------------
int processArguments(int argc, char* argv[], WORK_PARAMS **ppWP)
{
  int iErr, i, iV, *pI;
  DATA_FILE *pDF;
  WORK_PARAMS *pWP = NULL;

  // spracovanie a kontrola vstupnych parametrov
  do {
    *ppWP = NULL;
    iErr = allocBuffer((void *) &pWP, sizeof(WORK_PARAMS));
    if (iErr != ERR_OK)
      break;
    *ppWP = pWP;

    // inicializacia niektorych pracovnych parametrov
    pWP->iCM = CM_V2A;
    pWP->iMM = 0xFF;
    pWP->iMif85 = 0xFF;
    pWP->iVPI = 0xFF;
    pWP->iAF = -1;
    pWP->iBC = -1;
    pWP->iDE = -1;
    pWP->iHL = -1;
    pWP->iPC = -1;
    pWP->iSP = -1;

    for (i = 1; i < argc; i++) {
      if (strcmp(argv[i], "-h") == 0) {
        iErr = ERR_PRINT_USAGE;
        break;
      }
      else if (strcmp(argv[i], "-vb") == 0) {
        pWP->bVB = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-pack") == 0) {
        pWP->bPack = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-allram") == 0) {
        pWP->bAllram = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-ei") == 0) {
        pWP->bEI = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-vpi") == 0) {
        if (i + 1 >= argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i], NULL);
          break;
        }
        else {
          i++;
          // zakazanie/povolenie prerusenia od Video CPU
          iV = parseNumber(argv[i]);
          if (iV < 0 || iV > 1) {
            iErr = printError(ERR_BAD_VALUE, argv[i - 1], argv[i]);
            break;
          }
          pWP->iVPI = iV;
        }
        continue;
      }
      else if (strcmp(argv[i], "-mus") == 0) {
        pWP->bMus = TRUE;
        continue;
      }
      else if (strcmp(argv[i], "-mm") == 0) {
        if (i + 1 >= argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i], NULL);
          break;
        }
        else {
          i++;
          // mapovanie pamati 256kB
          iV = parseNumber(argv[i]);
          if (iV < 0 || iV > 127) {
            iErr = printError(ERR_BAD_MM_VALUE, argv[i], NULL);
            break;
          }
          pWP->iMM = iV;
        }
        continue;
      }
      else if (strcmp(argv[i], "-mif85") == 0) {
        if (i + 1 >= argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i], NULL);
          break;
        }
        else {
          i++;
          // zakazanie/povolenie prerusenia od CT0
          iV = parseNumber(argv[i]);
          if (iV < 0 || iV > 1) {
            iErr = printError(ERR_BAD_VALUE, argv[i - 1], argv[i]);
            break;
          }
          pWP->iMif85 = iV;
        }
        continue;
      }
      else if (strcmp(argv[i], "-cm") == 0) {
        if (i + 1 == argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i], NULL);
          break;
        }
        else {
          i++;
          // V1/V2/V2A/V3/MATO/ALFA/ALFA2/C2717
          if (stricmp(argv[i], "V1") == 0)
            pWP->iCM = CM_V1;
          else if (stricmp(argv[i], "V2") == 0)
            pWP->iCM = CM_V2;
          else if (stricmp(argv[i], "V2A") == 0)
            pWP->iCM = CM_V2A;
          else if (stricmp(argv[i], "V3") == 0)
            pWP->iCM = CM_V3;
          else if (stricmp(argv[i], "MATO") == 0)
            pWP->iCM = CM_MATO;
          else if (stricmp(argv[i], "ALFA") == 0)
            pWP->iCM = CM_ALFA;
          else if (stricmp(argv[i], "ALFA2") == 0)
            pWP->iCM = CM_ALFA2;
          else if (stricmp(argv[i], "C2717") == 0)
            pWP->iCM = CM_C2717;
          else {
            iErr = printError(ERR_BAD_COMP_MODEL, argv[i], NULL);
            break;
          }
        }
        continue;
      }
      else if (strcmp(argv[i], "-reg") == 0) {
        if (i + 2 >= argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i], NULL);
          break;
        }
        else {
          i++;
          // AF/BC/DE/HL/PC/SP
          if (stricmp(argv[i], "AF") == 0)
            pI = &pWP->iAF;
          else if (stricmp(argv[i], "BC") == 0)
            pI = &pWP->iBC;
          else if (stricmp(argv[i], "DE") == 0)
            pI = &pWP->iDE;
          else if (stricmp(argv[i], "HL") == 0)
            pI = &pWP->iHL;
          else if (stricmp(argv[i], "PC") == 0)
            pI = &pWP->iPC;
          else if (stricmp(argv[i], "SP") == 0)
            pI = &pWP->iSP;
          else {
            iErr = printError(ERR_BAD_REG_NAME, argv[i], NULL);
            break;
          }
          i++;
          // hodnota
          iV = parseNumber(argv[i]);
          if (iV < 0 || iV > 65535) {
            iErr = printError(ERR_BAD_ADDRESS, argv[i - 2], &iV);
            break;
          }
          *pI = iV;
        }
        continue;
      }
      else if (strcmp(argv[i], "-pg") == 0) {
        if (i + 2 >= argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i], NULL);
          break;
        }
        else {
          i++;
          // cislo stranky
          iV = parseNumber(argv[i]);
          if (iV < 0 || iV > 15) {
            iErr = printError(ERR_BAD_PAGE_NUM, argv[i], NULL);
            break;
          }
          i++;
          // struktura DATA_FILE, meno suboru, stranka
          iErr = addDataFile(pWP, argv[i], iV, -1);
          if (iErr != ERR_OK)
            break;
        }
        continue;
      }
      else if (strcmp(argv[i], "-mem") == 0) {
        if (i + 2 >= argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i], NULL);
          break;
        }
        else {
          i++;
          // adresa
          iV = parseNumber(argv[i]);
          if (iV < 0 || iV > 65535) {
            iErr = printError(ERR_BAD_ADDRESS, argv[i - 1], &iV);
            break;
          }
          i++;
          // struktura DATA_FILE, meno suboru, adresa
          iErr = addDataFile(pWP, argv[i], -1, iV);
          if (iErr != ERR_OK)
            break;
        }
        continue;
      }
      else if (strcmp(argv[i], "-hex") == 0) {
        if (i + 2 >= argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i], NULL);
          break;
        }
        else {
          i++;
          // adresa
          iV = parseNumber(argv[i]);
          if (iV < 0 || iV > 65535) {
            iErr = printError(ERR_BAD_ADDRESS, argv[i - 1], &iV);
            break;
          }
          i++;
          // struktura DATA_FILE, hexdata string, adresa
          iErr = addDataFile(pWP, argv[i], -2, iV);
          if (iErr != ERR_OK)
            break;
        }
        continue;
      }
      else if (strcmp(argv[i], "-o") == 0) {
        if (i + 1 == argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i], NULL);
          break;
        }
        i++;
        if (pWP->pszFnOut != NULL) {
          iErr = printError(ERR_OUTPUT_FILE_MORE, pWP->pszFnOut, argv[i]);
          break;
        }
        pWP->pszFnOut = argv[i];
        continue;
      }
      else if (strcmp(argv[i], "-rom") == 0) {
        if (i + 1 == argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i], NULL);
          break;
        }
        i++;
        if (pWP->pszFnRom != NULL) {
          iErr = printError(ERR_ROM_FILE_MORE, pWP->pszFnRom, argv[i]);
          break;
        }
        pWP->pszFnRom = argv[i];
        continue;
      }
      else if (strcmp(argv[i], "-psn") == 0) {
        if (i + 1 == argc) {
          iErr = printError(ERR_NO_OPTION_VALUE, argv[i], NULL);
          break;
        }
        i++;
        if (pWP->pszFnPsn != NULL) {
          iErr = printError(ERR_PSN_FILE_MORE, pWP->pszFnPsn, argv[i]);
          break;
        }
        pWP->pszFnPsn = argv[i];
        continue;
      }
    }
    if (iErr != ERR_OK)
      break;

    // vystupny subor musi byt uvedeny
    if (pWP->pszFnOut == NULL) {
      iErr = printError(ERR_NO_OUTPUT_FILE, NULL, NULL);
      break;
    }

    // musi byt uvedeny aspon jeden vstupny subor alebo zdrojovy PSN
    if (pWP->pszFnPsn == NULL && pWP->pDFfirst == NULL) {
      iErr = printError(ERR_NO_INPUT_FILES, NULL, NULL);
      break;
    }

    // zisti, ci sa pripaja blok do stranok 4 az 15
    // ak ano, zapne sa rozsirenie 256kB
    if (pWP->iMM == 0xFF) {
      iV = FALSE;
      pDF = pWP->pDFfirst;
      while (pDF) {
        if (pDF->iPage >= 4) {
          iV = TRUE;
          break;
        }
        pDF = pDF->pNext;
      }
      if (iV)
        pWP->iMM = 0;
    }

    // kontrola dovolenych kombinacii volieb
    if (pWP->iMif85 != 0xFF && pWP->bMus)
      iErr = printError(ERR_BAD_OPTIONS, "-mif85", "-mus");
    if (pWP->bAllram && pWP->iCM != CM_V2A && pWP->iCM != CM_V3 && pWP->iCM != CM_C2717)
      iErr = printError(ERR_BAD_OPTIONS, "-allram", "-cm");
    if ((pWP->iMM != 0xFF || pWP->iVPI != 0xFF) && pWP->iCM != CM_V2A && pWP->iCM != CM_V3)
      iErr = printError(ERR_BAD_OPTIONS, "-mm -vpi", "-cm");

  } while (FALSE);

  return iErr;
}
//---------------------------------------------------------------------------
int fillMemoryPages(WORK_PARAMS *pWP)
{
  int iErr, iLen, iPage, iAdr, iSize;
  DATA_FILE *pDF;
  BYTE *pbt;
  char *pc, c;

  // alokuj buffer
  iErr = allocBuffer((void *) &pWP->pbtBuff, BUFFER_SIZE);
  if (iErr != ERR_OK)
    return iErr;

  pDF = pWP->pDFfirst;
  while (pDF) {
    if (pDF->iAddress >= 0 && pDF->iPage == -2) { // hexdata retazec
      if (pWP->bVB)
        printf("Parsing hexdata string `%s` and storing from address %d\n", pDF->pszFn, pDF->iAddress);
      if (strlen(pDF->pszFn) % 2 == 1) {
        iErr = printError(ERR_INVALID_HEXSTRING, pDF->pszFn, NULL);
        break;
      }
      if (pDF->iAddress + strlen(pDF->pszFn) / 2 > 65536) {
        iErr = printError(ERR_HEXSTRING_TO_LONG, pDF->pszFn, &pDF->iAddress);
        break;
      }

      iLen = 0;
      pbt = pWP->pbtBuff;         // zdrojova adresa
      pc = pDF->pszFn;            // prvy znak hexdata retazca
      while (*pc) {
        c = *pc++;
        if (!isxdigit(c)) {
          iErr = printError(ERR_INVALID_HEXSTRING, pDF->pszFn, NULL);
          break;
        }
        if (c >= 'a')
          c = c - 'a' + 10;
        else if (c >= 'A')
          c = c - 'A' + 10;
        else
          c = c - '0';
        iPage = c * 16;

        c = *pc++;
        if (!isxdigit(c)) {
          iErr = printError(ERR_INVALID_HEXSTRING, pDF->pszFn, NULL);
          break;
        }
        if (c >= 'a')
          c = c - 'a' + 10;
        else if (c >= 'A')
          c = c - 'A' + 10;
        else
          c = c - '0';
        iPage += c;

        *pbt++ = (BYTE) iPage;
        iLen++;
      }
      if (iErr < ERR_OK)
        break;
    }

    if (pDF->iAddress >= 0) {
      iAdr = pDF->iAddress;
      if (pDF->iPage != -2) {
        // obsah suboru do pracovneho buffra
        if (pWP->bVB)
          printf("Loading file `%s` from address %d\n", pDF->pszFn, iAdr);
        iErr = readFile(pDF->pszFn, pWP->pbtBuff, BUFFER_SIZE - iAdr); // iErr = dlzka suboru
        if (iErr < ERR_OK)
          break;
        iLen = iErr;                // celkova dlzka
      }
      iPage = iAdr / PAGE_SIZE;   // cislo prvej stranky
      iAdr = iAdr % PAGE_SIZE;    // offset v prvej stranke
      iSize = PAGE_SIZE - iAdr;   // max. pocet bytov do prvej stranky
      pbt = pWP->pbtBuff;         // zdrojova adresa
      while (iLen > 0) {
        // alokovanie pamati pre danu stranku
        iErr = allocBuffer((void *) &pWP->pbtaPages[iPage], PAGE_SIZE);
        if (iErr != ERR_OK)
          break;
        pWP->iaPgLen[iPage] = PAGE_SIZE;

        if (iSize > iLen)
          iSize = iLen;
        memcpy(pWP->pbtaPages[iPage] + iAdr, pbt, iSize);
        pbt += iSize;             // posun zdrojovej adresy
        iLen = iLen - iSize;      // odpocet zostavajucej dlzky
        iSize = PAGE_SIZE;        // maximum pre dalsiu stranku
        iAdr = 0;                 // dalsia stranka od offsetu 0
        iPage++;                  // dalsia stranka
      }
      if (iErr < ERR_OK)
        break;
    }
    else if (pDF->iPage >= 0) {
      // alokovanie pamati pre danu stranku
      iErr = allocBuffer((void *) &pWP->pbtaPages[pDF->iPage], PAGE_SIZE);
      if (iErr != ERR_OK)
        break;
      pWP->iaPgLen[pDF->iPage] = PAGE_SIZE;

      if (pWP->bVB)
        printf("Loading file `%s` into page %d\n", pDF->pszFn, pDF->iPage);
      iErr = readFile(pDF->pszFn, pWP->pbtaPages[pDF->iPage], PAGE_SIZE);
      if (iErr < ERR_OK)
        break;
      iErr = ERR_OK; // vynuluj chybu (dolezite!)
    }

    pDF = pDF->pNext;
  }

  if (pWP->pszFnRom) {
    if (pWP->bVB)
      printf("Loading ROM file `%s`\n", pWP->pszFnRom);
    memset(pWP->pbtBuff, 0xFF, PAGE_SIZE);
    iErr = readFile(pWP->pszFnRom, pWP->pbtBuff, PAGE_SIZE); // iErr = dlzka suboru
    if (iErr > 0) {
      iLen = (iErr + 1023) & 0xFC00; // zarovnanie na kB
      iErr = allocBuffer((void *) &pWP->pbtRom, iLen);
      if (iErr == ERR_OK) {
        memcpy(pWP->pbtRom, pWP->pbtBuff, iLen);
        pWP->iRomLen = iLen;
      }
    }
  }

  return iErr;
}
//---------------------------------------------------------------------------
int addDataFile(WORK_PARAMS *pWP, char *pszStr, int iPage, int iAddr)
{
  DATA_FILE *pDF;
  int iErr;

  pDF = NULL; // dolezite !!!
  iErr = allocBuffer((void *) &pDF, sizeof(DATA_FILE));
  if (iErr != ERR_OK)
    return iErr;;
  pDF->pszFn = pszStr;
  pDF->iPage = iPage;
  pDF->iAddress = iAddr;
  pDF->pNext = NULL;
  if (pWP->pDFfirst) {
    pWP->pDFlast->pNext = pDF;
    pWP->pDFlast = pDF;
  }
  else {
    pWP->pDFfirst = pDF;
    pWP->pDFlast = pDF;
  }

  return ERR_OK;
}
//---------------------------------------------------------------------------
int allocBuffer(void **ppvBuff, int iSize)
{
  if (*ppvBuff)
    return ERR_OK;

  *ppvBuff = (void *) malloc(iSize);
  if (!*ppvBuff)
    return printError(ERR_OUT_OF_MEMORY, NULL, NULL);

  memset(*ppvBuff, 0, iSize);
  return ERR_OK;
}
//---------------------------------------------------------------------------
int parseNumber(char *pszStr)
{
  char *p;
  int iVal = strtol(pszStr, &p, 0);
  if (iVal < 0 || (iVal == 0 && *p != 0))
    iVal = -1;
  return iVal;
}
//---------------------------------------------------------------------------
int readFile(char *pszFileName, BYTE *pbtBuff, int iMaxLen)
{
  int iErr = ERR_OK, iLen;
  FILE *f;

  do {
    f = fopen(pszFileName, "rb");
    if (f == NULL) {
      iErr = printError(ERR_OPEN_SRC_FILE, pszFileName, NULL);
      break;
    }
    if (fseek(f, 0L, SEEK_END) != 0) {
      iErr = printError(ERR_ACCESS_SRC_FILE, pszFileName, NULL);
      break;
    }
    iLen = ftell(f);
    if (iLen < 0) {
      iErr = printError(ERR_LENGTH_SRC_FILE, pszFileName, NULL);
      break;
    }
    if (iLen > iMaxLen) {
      iErr = printError(ERR_SRC_FILE_TO_LONG, pszFileName, &iMaxLen);
      break;
    }
    if (fseek(f, 0L, SEEK_SET) != 0) {
      iErr = printError(ERR_ACCESS_SRC_FILE, pszFileName, NULL);
      break;
    }
    if (fread(pbtBuff, 1, iLen, f) != (size_t) iLen) {
      iErr = printError(ERR_READ_SRC_FILE, pszFileName, NULL);
      break;
    }
  } while (FALSE);

  if (f != NULL)
    fclose(f);

  if (iErr < ERR_OK)
    return iErr;

  return iLen;
}
//---------------------------------------------------------------------------
int writeFile(char *pszFileName, BYTE *pbtBuff, int iLen, BOOL bAppend)
{
  int iErr = ERR_OK;
  FILE *f;

  if (iLen <= 0)
    return iErr;

  do {
    if (bAppend)
      f = fopen(pszFileName, "ab");
    else
      f = fopen(pszFileName, "wb");
    if (f == NULL) {
      iErr = printError(ERR_OPEN_DEST_FILE, pszFileName, NULL);
      break;
    }
    if (!bAppend && fseek(f, 0L, SEEK_SET) != 0) {
      iErr = printError(ERR_ACCESS_DEST_FILE, pszFileName, NULL);
      break;
    }
    if (fwrite(pbtBuff, 1, iLen, f) != (size_t) iLen) {
      iErr = printError(ERR_WRITE_DEST_FILE, pszFileName, NULL);
      break;
    }
  } while (FALSE);

  if (f != NULL)
    fclose(f);

  return iErr;
}
//---------------------------------------------------------------------------
// Metoda sa pokusi spakovat blok o dlzke 'iLen' zo 'pbtSrc' do 'pbtDest'.
// 'pbtSrc' a 'pbtDest' musia byt osobitne alokovane bloky. Oba musia mat velkost
// minimalne 'iLen'.
// Navratovou hodnotou je dlzka spakovaneho bloku v 'pbtDest'.
// Ak je tato hodnota = 1, zdrojovy blok bol vyplneny jednou hodnotou, ktora je
// ulozena v '*pbtDest'. Ak je vratena hodnota = 'iLen', blok sa nepodarilo spakovat
// a do 'pbtDest' je skopirovany povodny blok zo 'pbtSrc'. V ostatnych pripadoch
// (navratova hodnota < 'iLen') je blok spakovany jednoduchou RLE kompresiou.
// Format kompresovanych dat: Flag Data Flag Data Flag Data ...
// Flag - 0x00 az 0x7F => 1 byte ako Data pre 3 az 130 rovnakych bytov
// Flag - 0x80 az 0xFF => blok 1 az 128 nepakovatelnych bytov ako Data
//---------------------------------------------------------------------------
int packBlock(BYTE *pbtDest, BYTE *pbtSrc, int iLen)
{
  BYTE *pbtS = pbtSrc;
  BYTE *pbtMaxS = pbtSrc + iLen;
  BYTE *pbtD = pbtDest;
  BYTE *pbtMaxD = pbtDest + iLen - 1;
  BYTE btX, btCnt = 0;
  int iPackedLen;

  // zistenie, ci je blok vyplneny iba jednou hodnotou
  btX = *pbtSrc;
  while (++pbtS < pbtMaxS) {
    if (btX != *pbtS)
      break;
  }
  if (pbtS == pbtMaxS) {
    *pbtDest = btX;  // blok je vyplneny iba jednou hodnotou
    return 1;
  }

  // kompresia bloku
  pbtS = pbtSrc;
  while (pbtS < pbtMaxS && pbtD < pbtMaxD) {
    btX = *pbtS++;
    if (pbtMaxS - pbtS >= 2 && btX == *pbtS && btX == *(pbtS + 1)) {
      // nasli sa rovnake 3 byty za sebou
      if (btCnt > 0) { // ak tomu predchadzal nepakovatelny blok, zaznamenaj to
        *(pbtD - btCnt) = (BYTE) ((btCnt - 1) | 0x80); // pocet - flag
        pbtD++;
        btCnt = 0;
      }

      if (pbtD < pbtMaxD) {
        // najdenie dalsich opakujucich sa bytov
        pbtS += 2;
        while (pbtS < pbtMaxS && btX == *pbtS && btCnt < 127) { // max 130 (= 3 + 127)
          pbtS++;
          btCnt++;
        }
        *pbtD++ = btCnt; // pocet - flag
        *pbtD++ = btX;   // byte
        btCnt = 0;
      }
    }
    else {
      *(++pbtD) = btX;            // nepakovatelny byte
      if (++btCnt == 0x80) {      // max 128
        *(pbtD - btCnt) = 0xFF;   // pocet 128 - flag ((128 - 1) or 0x80)
        pbtD++;
        btCnt = 0;
      }
    }
  }
  if (btCnt > 0 && pbtD < pbtMaxD + 1) {
    *(pbtD - btCnt) = (BYTE) ((btCnt - 1) | 0x80);
    pbtD++;
  }

  // dlzka spakovaneho bloku
  iPackedLen = pbtD - pbtDest;
  if (iPackedLen < iLen && pbtS == pbtMaxS)
    return iPackedLen; // kompresia uspesna

  // neuspesna kompresia - do pbtDest sa ulozi povodny blok
  memcpy(pbtDest, pbtSrc, iLen);
  return iLen;
}
//---------------------------------------------------------------------------
// Metoda sa pokusi rozpakovat blok o dlzke 'iSrclen' zo 'pbtSrc' do 'pbtDest'.
// 'pbtSrc' a 'pbtDest' musia byt osobitne alokovane bloky.
// 'iDestlen' je dlzka cieloveho 'pbtDest' buffra. V pripade, ze 'iSrclen' = 1
// alebo 'iSrclen' = 'iDestlen', 'iDestlen' je zaroven pozadovana dlzka cielovych
// dat. Ak je 'iSrclen' = 1, tak sa bytom na '*pbtSrc' vyplni 'pbtDest' o dlzke
// 'iDestlen'. Ak je 'iSrclen' = 'iDestlen', jedna sa o nespakovany blok a prevedie
// sa iba prekopirovanie zo 'pbtSrc' do 'pbtDest'.
// Navratovou hodnotou je pocet bytov ulozenych v 'pbtDest'.
// Ak je vratena hodnota -1, su chybne vstupne parametre.
// Ak je vratena hodnota -2, doslo ku chybe pri rozpakovani, teda 'pbtSrc' data
// nezodpovedaju pakovanemu bloku, ktory sa po rozpakovani "vojde" do 'pbtDest'.
int unpackBlock(BYTE *pbtDest, int iDestlen, BYTE *pbtSrc, int iSrclen)
{
  BYTE *pbtOrigDest = pbtDest;
  BYTE *pbtMaxS = pbtSrc + iSrclen;
  BYTE *pbtMaxD = pbtDest + iDestlen;
  BYTE btX, btY;

  if (pbtDest == NULL || pbtSrc == NULL
      || iSrclen == 0 || iDestlen == 0 || iDestlen < iSrclen)
    return -1;

  if (iSrclen == 1) {
    memset(pbtDest, *pbtSrc, iDestlen);
    return iDestlen;
  }
  if (iSrclen == iDestlen) {
    memcpy(pbtDest, pbtSrc, iDestlen);
    return iDestlen;
  }

  while (pbtSrc < pbtMaxS && pbtDest < pbtMaxD) {
    btX = *pbtSrc++;
    if (btX & 0x80) { // nepakovane data 1 az 128 => 0x80 az 0xFF
      btX &= 0x7F;
      btX++;
      while (btX-- > 0 && pbtDest < pbtMaxD)
        *pbtDest++ = *pbtSrc++;
    }
    else {  // pakovane data 3 az 130 => 0x00 az 0x7F
      btX += (BYTE) 3;
      btY = *pbtSrc++;
      while (btX-- > 0 && pbtDest < pbtMaxD)
        *pbtDest++ = btY;
    }
  }

  if (pbtMaxS - pbtSrc > 0)
    return -2;

  return pbtDest - pbtOrigDest;
}
//---------------------------------------------------------------------------
int printError(int iErr, void *pvPar, void *pvPar2)
{
  switch (iErr) {
    case ERR_OUT_OF_MEMORY :
      fprintf(stderr, "Out of memory!");
      break;

    case ERR_BAD_OPTIONS :
      fprintf(stderr, "Bad combination of options: %s %s", pvPar, pvPar2);
      break;

    case ERR_NO_OPTION_VALUE :
      fprintf(stderr, "Missing option value(s) '%s'!", (char *) pvPar);
      break;

    case ERR_OPEN_SRC_FILE :
      fprintf(stderr, "Can't open source file '%s'!", (char *) pvPar);
      break;

    case ERR_ACCESS_SRC_FILE :
      fprintf(stderr, "Can't access source file '%s'!", (char *) pvPar);
      break;

    case ERR_LENGTH_SRC_FILE :
      fprintf(stderr, "Can't retrieve length of the source file '%s'!", (char *) pvPar);
      break;

    case ERR_SRC_FILE_TO_LONG :
      fprintf(stderr, "Source file '%s' is too long! Max length %d bytes.", (char *) pvPar, *(int *) pvPar2);
      break;

    case ERR_READ_SRC_FILE :
      fprintf(stderr, "Can't read from the source file '%s'!", (char *) pvPar);
      break;

    case ERR_OPEN_DEST_FILE :
      fprintf(stderr, "Can't open output file '%s'!", (char *) pvPar);
      break;

    case ERR_ACCESS_DEST_FILE :
      fprintf(stderr, "Can't access output file '%s'!", (char *) pvPar);
      break;

    case ERR_WRITE_DEST_FILE :
      fprintf(stderr, "Can't write to the output file '%s'!", (char *) pvPar);
      break;

    case ERR_BAD_ADDRESS :
      fprintf(stderr, "Value %d is out of range <0, 65535>: %s", *(int *) pvPar2, (char *) pvPar);
      break;

    case ERR_NO_INPUT_FILES :
      fprintf(stderr, "Input file(s) not specified!");
      break;

    case ERR_NO_OUTPUT_FILE :
      fprintf(stderr, "Output file not specified!");
      break;

    case ERR_OUTPUT_FILE_MORE :
      fprintf(stderr, "Only one output file can be specified! `%s` or `%s`?", (char *) pvPar, (char *) pvPar2);
      break;

    case ERR_ROM_FILE_MORE :
      fprintf(stderr, "Only one ROM file can be specified! `%s` or `%s`?", (char *) pvPar, (char *) pvPar2);
      break;

    case ERR_PSN_FILE_MORE :
      fprintf(stderr, "Only one PSN file can be specified! `%s` or `%s`?", (char *) pvPar, (char *) pvPar2);
      break;

    case ERR_BAD_COMP_MODEL :
      fprintf(stderr, "Bad computer model '%s'!", (char *) pvPar);
      break;

    case ERR_BAD_REG_NAME :
      fprintf(stderr, "Bad register name '%s'!", (char *) pvPar);
      break;

    case ERR_BAD_PAGE_NUM :
      fprintf(stderr, "Bad page number `%s`!", (char *) pvPar);
      break;

    case ERR_BAD_MM_VALUE :
      fprintf(stderr, "Bad Memory Mapper value '%s'!", (char *) pvPar);
      break;

    case ERR_BAD_VALUE :
      fprintf(stderr, "Bad option value: %s %s", (char *) pvPar, (char *) pvPar2);
      break;

    case ERR_SNAP_BAD_HEADER :
      fprintf(stderr, "Bad or no PSN header in file: '%s'!", (char *) pvPar);
      break;

    case ERR_SNAP_UNK_VERSION :
      fprintf(stderr, "Bad PSN version `%d` in file: '%s'!", (int) *(BYTE *) pvPar2, (char *) pvPar);
      break;

    case ERR_SNAP_UNK_MODEL :
      fprintf(stderr, "Unknown computer model `%d` in PSN file: '%s'!", *(int *) pvPar2, (char *) pvPar);
      break;

    case ERR_SNAP_BAD_MEM_BLK :
      fprintf(stderr, "Bad size of memory blocks after unpack in PSN file: '%s'!", (char *) pvPar);
      break;

    case ERR_INVALID_HEXSTRING :
      fprintf(stderr, "Invalid hexstring: '%s'!", (char *) pvPar);
      break;

    case ERR_HEXSTRING_TO_LONG :
      fprintf(stderr, "Bytes from hexstring '%s' do not fit into memory from address %d.", (char *) pvPar, *(int *) pvPar2);
      break;
  }

  fprintf(stderr, "\n\n");

  return iErr;
}
//---------------------------------------------------------------------------
void version(void)
{
  printf("%s v%d.%d - %s\n\n",
         PROG_NAME, PROG_VERSION / 16, PROG_VERSION % 16, PROG_AUTHOR);
}
//---------------------------------------------------------------------------
void pressEnter(void)
{
  printf("\nPress ENTER ...");
  while (getchar() != '\n')
    ;
}
//---------------------------------------------------------------------------
void usage(void)
{
  printf("Creates PSN Snapshot file for PMD 85 Emulator\n");
  printf("Usage: %s [options]\n", PROG_EXE_NAME);
  printf("\nOptions:\n");
  printf("  -o outfile.psn\toutput file name\n");
  printf("  -pack\t\t\tpack Snapshot content\n");
  printf("  -cm CM\t\tcomputer [V2A]; CM=V1/V2/V2A/V3/MATO/ALFA/ALFA2/C2717\n");
  printf("  -psn filename.psn\ttemplate PSN file\n");
  printf("  -rom filename.ext\tROM file\n");
  printf("  -mem adr filename.ext\tload file into memory from given address\n");
  printf("  -pg PG filename.ext\tload file into memory page; PG=<0,15>\n");
  printf("  -hex adr hexdata\tstore direct bytes into memory from given address\n");
  printf("  -reg RG value\t\tregister value; RG=AF/BC/DE/HL/PC/SP\n");
  printf("  -allram\t\tswitch AllRAM on\n");
  printf("  -ei\t\t\tenable global interrupt\n");
  printf("  -vpi intVCPU\t\tplug VideoCPU interrupt circuit; intVCPU=0/1 di/ei int\n");
  printf("  -mm value\t\t256kB memory mapping port value\n");
  printf("  -mif85 intCT0\t\tplug MIF 85; intCT0=0/1 dis/ena interrupt from CT0\n");
  printf("  -mus\t\t\tplug MUSICA interface\n");
  printf("  -vb\t\t\tverbose\n");
  printf("  -h\t\t\tusage information\n");

  pressEnter();

  printf("\nAll numeric parameters must be nonnegative integers and can be in three forms:\n");
  printf("  decimal (e.g. 12345)\n");
  printf("  octal (e.g. 01234)\n");
  printf("  hexadecimal (e.g. 0x1234)\n\n");
}
//---------------------------------------------------------------------------

