//---------------------------------------------------------------------------
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "common.h"

//---------------------------------------------------------------------------
void prepareBlockLengths(SNAP *pSnap, WORD *pwaBlockLengths[]);

//---------------------------------------------------------------------------
// Ulozenie Snapshotu do suboru.
int saveSnapshot(WORK_PARAMS *pWP)
{
  int ii, iErr, iRamBlocks, iLen;
  WORD *pwaBlockLengths[16];
  SNAP *pSnap = pWP->pSnap;

  // verzia 1
  BYTE btVer = 1;
  WORD wBlkOff = SNAP_HEADER_V1_LEN;
#if (SNAP_VERSION > 1)
  if (pWP->bMus || pWP->iMif85 != 0xFF || pWP->iVPI != 0xFF || pWP->iMM != 0xFF) {
    // verzia 2
    btVer = 2;
    wBlkOff = SNAP_HEADER_V2_LEN;
  }
#endif

  // priprava Snapshotu
  if (pWP->pszFnPsn == NULL) {
    // nastav "default" hodnoty
    memcpy(pSnap->cSignature, SNAP_SIGNATURE, SNAP_SIGNATURE_LEN);
    if (pWP->iAF == -1)
      pWP->iAF = 0x0046;
    if (pWP->iBC == -1)
      pWP->iBC = 0;
    if (pWP->iDE == -1)
      pWP->iDE = 0;
    if (pWP->iHL == -1)
      pWP->iHL = 0;
    if (pWP->iPC == -1)
      pWP->iPC = 0x8000;
    if (pWP->iSP == -1)
      pWP->iSP = 0x8000;

    pSnap->btSystemPioCwr = 0x8A;
    pSnap->btSystemPioC = 0;
    pSnap->btSystemPioA = 0;

    pSnap->btGpioCwr = 0xAD;
    pSnap->btGpioC = 0x82;
    pSnap->btGpioB = 0;
    pSnap->btGpioA = 0;
    pSnap->btGpioInt = 6;

    pSnap->btIms2Cwr = 0x9B;
    pSnap->btIms2C = 0;
    pSnap->btIms2B = 0;
    pSnap->btIms2A = 0;
    pSnap->btIms2Int = 0;

    pSnap->btTimer0Cwr = 0x32;
    pSnap->btTimer0L = 0xFF;
    pSnap->btTimer0H = 0xFF;
    pSnap->btTimer1Cwr = 0x76;
    pSnap->btTimer1L = 0x20;
    pSnap->btTimer1H = 0;
    pSnap->btTimer2Cwr = 0xB2;
    pSnap->btTimer2L = 0xFF;
    pSnap->btTimer2H = 0xFF;

    pSnap->btUsartCwr = 0xED;
    pSnap->btUsartSync1 = 0;
    pSnap->btUsartSync2 = 0;
    pSnap->btUsartCmd = 0x23;
  }

  pSnap->btVersion = btVer;
  pSnap->wBlocksOffset = wBlkOff;
  pSnap->btModel = (BYTE) pWP->iCM;
  pSnap->btInterrupt = (BYTE) ((pWP->bEI) ? 1 : 0);
  if (pWP->bAllram)
    pSnap->btSystemPioCwr = 0x82;

  if (pWP->iAF != -1)
    pSnap->wAF = (WORD) ((pWP->iAF & 0xFFD7) | 0x02);
  if (pWP->iBC != -1)
    pSnap->wBC = (WORD) pWP->iBC;
  if (pWP->iDE != -1)
    pSnap->wDE = (WORD) pWP->iDE;
  if (pWP->iHL != -1)
    pSnap->wHL = (WORD) pWP->iHL;
  if (pWP->iPC != -1)
    pSnap->wPC = (WORD) pWP->iPC;
  if (pWP->iSP != -1)
    pSnap->wSP = (WORD) pWP->iSP;

#if (SNAP_VERSION > 1)
  if (btVer > 1) {
    pSnap->btVideoIntVGA = (BYTE) ((pWP->iVPI == 0xFF) ? 0xFF : ((pWP->iVPI) ? 0x87 : 0x07));
    pSnap->btMemMap = (BYTE) pWP->iMM;
    pSnap->btMif85Int = (BYTE) pWP->iMif85;
    memset(pSnap->btaMif85Regs, 0, 32);
    if (pWP->bMus) {
      pSnap->btMusicaTimer0Cwr = 0x32;
      pSnap->btMusicaTimer0L = 0xFF;
      pSnap->btMusicaTimer0H = 0xFF;
      pSnap->btMusicaTimer1Cwr = 0x72;
      pSnap->btMusicaTimer1L = 0xFF;
      pSnap->btMusicaTimer1H = 0xFF;
      pSnap->btMusicaTimer2Cwr = 0xB2;
      pSnap->btMusicaTimer2L = 0xFF;
      pSnap->btMusicaTimer2H = 0xFF;
    }
    else {
      pSnap->btMusicaTimer0Cwr = 0xFF;
      pSnap->btMusicaTimer0L = 0;
      pSnap->btMusicaTimer0H = 0;
      pSnap->btMusicaTimer1Cwr = 0;
      pSnap->btMusicaTimer1L = 0;
      pSnap->btMusicaTimer1H = 0;
      pSnap->btMusicaTimer2Cwr = 0;
      pSnap->btMusicaTimer2L = 0;
      pSnap->btMusicaTimer2H = 0;
    }
  }
#endif // (SNAP_VERSION > 1)

  // priprava ROM - Monitora
  if (pWP->pbtRom) {
    if (pWP->iCM == CM_V3 && pWP->iRomLen > 8192)
      pWP->iRomLen = 8192;
    if (pWP->bPack) {
      iLen = packBlock(pWP->pbtBuff, pWP->pbtRom, pWP->iRomLen);
      memcpy(pWP->pbtRom, pWP->pbtBuff, iLen);
    }
    else
      iLen = pWP->iRomLen;
    if (iLen == pWP->iRomLen)
      pSnap->wRomLength = (WORD) (iLen | 0x8000);
    else
      pSnap->wRomLength = (WORD) iLen;
  }

  // * priprava jednotlivych blokov *
#if (SNAP_VERSION > 1)
  iRamBlocks = (btVer > 1 && pWP->iMM != 0xFF) ? 16 : 4;
#else
  iRamBlocks = 4;
#endif

  // adresy poloziek dlziek jednotlivych blokov v strukture Snap
  prepareBlockLengths(pSnap, pwaBlockLengths);

  for (ii = 0; ii < iRamBlocks; ii++) {
    pWP->iaPgLen[ii] = 0;
    *pwaBlockLengths[ii] = 0;

    if (ii == 2 && !(pWP->iCM == CM_V2A || pWP->iCM == CM_V3 || pWP->iCM == CM_C2717))
      continue;

    iLen = 0;
    if (pWP->pbtaPages[ii]) {
      if (pWP->bPack) {
        iLen = packBlock(pWP->pbtBuff, pWP->pbtaPages[ii], SNAP_BLOCK_LEN);
        memcpy(pWP->pbtaPages[ii], pWP->pbtBuff, iLen);
      }
      else
        iLen = SNAP_BLOCK_LEN;
    }
    pWP->iaPgLen[ii] = iLen;
    *pwaBlockLengths[ii] = (WORD) iLen;
  }

  // * zapis Snapshotu do suboru *
  do {
    // hlavicka Snapshotu
    iLen = wBlkOff;
    iErr = writeFile(pWP->pszFnOut, (BYTE *) pSnap, iLen, FALSE);
    if (iErr != ERR_OK)
      break;

    // ROM Monitor
    if (pWP->pbtRom) {
      iLen = (int) (pSnap->wRomLength & 0x7FFF);
      iErr = writeFile(pWP->pszFnOut, pWP->pbtRom, iLen, TRUE);
      if (iErr != ERR_OK)
        break;
    }

    for (ii = 0; ii < iRamBlocks; ii++) {
      if (ii == 2 && !(pWP->iCM == CM_V2A || pWP->iCM == CM_V3 || pWP->iCM == CM_C2717))
        continue;

      iLen = (int) *pwaBlockLengths[ii];
      if (iLen > 0 && pWP->pbtaPages[ii]) {
        iErr = writeFile(pWP->pszFnOut, pWP->pbtaPages[ii], iLen, TRUE);
        if (iErr != ERR_OK)
          break;
      }
    }

  } while (FALSE);

  return iErr;
}
//---------------------------------------------------------------------------
// Precitanie Snapshotu zo suboru.
int loadSnapshot(WORK_PARAMS *pWP)
{
  BYTE *pbtSrc = NULL;
  BYTE *pbtDest = NULL;
  int iErr = ERR_OK;
  int iHeaderLen, iModel, iRamBlocks;
  int iLen, iLenX, ii;
  FILE *f;
  WORD *pwaBlockLengths[16];
  SNAP *pSnap = pWP->pSnap;

  do {
    f = fopen(pWP->pszFnPsn, "rb");
    if (f == NULL) {
      iErr = printError(ERR_OPEN_SRC_FILE, pWP->pszFnPsn, NULL);
      break;
    }

    if (fseek(f, 0L, SEEK_SET) != 0) {
      iErr = printError(ERR_ACCESS_SRC_FILE, pWP->pszFnPsn, NULL);
      break;
    }

    // precitaj zaciatok hlavicky Snapshotu zo suboru
    if (fread(pSnap, 1, 7, f) != (size_t) 7) {
      iErr = printError(ERR_READ_SRC_FILE, pWP->pszFnPsn, NULL);
      break;
    }

    // kontrola signatury
    if (memcmp(pSnap->cSignature, SNAP_SIGNATURE, SNAP_SIGNATURE_LEN) != 0) {
      iErr = printError(ERR_SNAP_BAD_HEADER, pWP->pszFnPsn, NULL);
      break;
    }

    // kontrola verzie
    if (pSnap->btVersion == 0 || pSnap->btVersion > SNAP_VERSION) {
      iErr = printError(ERR_SNAP_UNK_VERSION, pWP->pszFnPsn, &pSnap->btVersion);
      break;
    }

    // kontrola velkosti hlavicky
    iHeaderLen = (int) pSnap->wBlocksOffset;
    if (iHeaderLen < SNAP_MIN_HEADER_LEN || iHeaderLen > SNAP_MAX_HEADER_LEN) {
      iErr = printError(ERR_SNAP_BAD_HEADER, pWP->pszFnPsn, NULL);
      break;
    }

    // kontrola modelu
    iModel = (int) pSnap->btModel;
    if (iModel > CM_LAST) {
      iErr = printError(ERR_SNAP_UNK_MODEL, pWP->pszFnPsn, &iModel);
      break;
    }

    // posun v subore opat na zaciatok Snapshotu
    if (fseek(f, 0L, SEEK_SET) != 0) {
      iErr = printError(ERR_ACCESS_SRC_FILE, pWP->pszFnPsn, NULL);
      break;
    }

    // precitaj celu hlavicku Snapshotu zo suboru
    if (fread(pSnap, 1, iHeaderLen, f) != (size_t) iHeaderLen) {
      iErr = printError(ERR_READ_SRC_FILE, pWP->pszFnPsn, NULL);
      break;
    }

    // * priprava jednotlivych blokov *
#if (SNAP_VERSION > 1)
    iRamBlocks = (pSnap->btVersion > 1) ? 16 : 4;
#else
    iRamBlocks = 4;
#endif

    // adresy poloziek dlziek jednotlivych blokov v strukture Snap
    prepareBlockLengths(pSnap, pwaBlockLengths);

    // kontrola velkosti blokov
    if ((pSnap->wRomLength & 0x7FFF) > SNAP_BLOCK_LEN) {
      iErr = printError(ERR_SNAP_BAD_HEADER, pWP->pszFnPsn, NULL);
      break;
    }
    for (ii = 0; ii < iRamBlocks; ii++) {
      if (*pwaBlockLengths[ii] > SNAP_BLOCK_LEN) {
        iErr = printError(ERR_SNAP_BAD_HEADER, pWP->pszFnPsn, NULL);
        break;
      }
    }
    if (iErr < 0)
      break;

    iErr = allocBuffer((void *) &pbtSrc, SNAP_BLOCK_LEN);
    if (iErr < 0)
      break;
    iErr = allocBuffer((void *) &pbtDest, SNAP_BLOCK_LEN);
    if (iErr < 0)
      break;

    // spracovanie bloku ROM Monitora
    iLenX = (int) pSnap->wRomLength;
    if (iLenX > 0) {
      iLen = iLenX & 0x7FFF; // iLenX.b15 = 1, ak je ROM v cistom tvare

      if (fread(pbtSrc, 1, iLen, f) != (size_t) iLen) {
        iErr = printError(ERR_READ_SRC_FILE, pWP->pszFnPsn, NULL);
        break;
      }

      iLenX = unpackBlock(pbtDest, (iLenX & 0x8000) ? iLen : SNAP_BLOCK_LEN, pbtSrc, iLen);
      if (iLenX < 0) {
        iErr = printError(ERR_SNAP_BAD_MEM_BLK, pWP->pszFnPsn, NULL);
        break;
      }

      if (pWP->pbtRom != NULL) {
        free(pWP->pbtRom);
        pWP->pbtRom = NULL;
      }
      iErr = allocBuffer((void *) &pWP->pbtRom, iLenX);
      if (iErr < 0)
        break;
      memcpy(pWP->pbtRom, pbtDest, iLenX);
      pWP->iRomLen = iLenX;
    }

    for (ii = 0; ii < iRamBlocks; ii++) {
      if (ii == 2 && !(iModel == CM_V2A || iModel == CM_V3 || iModel == CM_C2717))
        continue;

      iLen = (int) *pwaBlockLengths[ii];
      if (iLen == 0)
        continue;

      if (fread(pbtSrc, 1, iLen, f) != (size_t) iLen) {
        iErr = printError(ERR_READ_SRC_FILE, pWP->pszFnPsn, NULL);
        break;
      }

      iLen = unpackBlock(pbtDest, SNAP_BLOCK_LEN, pbtSrc, iLen);
      if (iLen < 0 || iLen != SNAP_BLOCK_LEN) {
        iErr = printError(ERR_SNAP_BAD_MEM_BLK, pWP->pszFnPsn, NULL);
        break;
      }

      iErr = allocBuffer((void *) &pWP->pbtaPages[ii], iLen);
      if (iErr < 0)
        break;
      memcpy(pWP->pbtaPages[ii], pbtDest, iLen);
      pWP->iaPgLen[ii] = iLen;
    }

  } while (FALSE);

  if (f != NULL)
    fclose(f);

  if (pbtSrc != NULL)
    free(pbtSrc);
  if (pbtDest != NULL)
    free(pbtDest);

  return iErr;
}
//---------------------------------------------------------------------------
void prepareBlockLengths(SNAP *pSnap, WORD *pwaBlockLengths[])
{
  // adresy poloziek s dlzkami jednotlivych blokov v strukture Snap
  pwaBlockLengths[0] = &pSnap->wBlock0Length;
  pwaBlockLengths[1] = &pSnap->wBlock1Length;
  pwaBlockLengths[2] = &pSnap->wBlock2Length;
  pwaBlockLengths[3] = &pSnap->wBlock3Length;
#if (SNAP_VERSION > 1)
  if (pSnap->btVersion > 1) {
    pwaBlockLengths[4] = &pSnap->wBlock4Length;
    pwaBlockLengths[5] = &pSnap->wBlock5Length;
    pwaBlockLengths[6] = &pSnap->wBlock6Length;
    pwaBlockLengths[7] = &pSnap->wBlock7Length;
    pwaBlockLengths[8] = &pSnap->wBlock8Length;
    pwaBlockLengths[9] = &pSnap->wBlock9Length;
    pwaBlockLengths[10] = &pSnap->wBlock10Length;
    pwaBlockLengths[11] = &pSnap->wBlock11Length;
    pwaBlockLengths[12] = &pSnap->wBlock12Length;
    pwaBlockLengths[13] = &pSnap->wBlock13Length;
    pwaBlockLengths[14] = &pSnap->wBlock14Length;
    pwaBlockLengths[15] = &pSnap->wBlock15Length;
  }
#endif // (SNAP_VERSION > 1)
}
//---------------------------------------------------------------------------

